#! /usr/bin/perl

    #               Passport Photo Maker

    #  Designed and implemented by John Walker in August 2004

    #  Author's home page: http://www.fourmilab.ch/.
    #  Program home page:  http://www.fourmilab.ch/netpbm/PassportPhoto/

    #   Create a passport photo image with as many copies as fit of
    #   a photograph with a given resolution on a specified paper
    #   size.

    #   When the image is printed, you should specify "bleed" mode--the
    #   image should extend to the edge of the print paper.  Otherwise, a
    #   border will be added and the images will be smaller than you
    #   requested.

=head1 NAME

passport_photo - Make a page of passport photos, ready to print

=head1 SYNOPSIS

B<passport_photo>
[B<--black>]
[B<--copyright>]
[B<--grey>/B<--gray>]
[B<--help>]
[B<--ifilter> I<program>]
[B<--inputfile> I<fname>]
[B<--ofilter> I<program>]
[B<--outputfile> I<fname>]
[B<--paperheight> I<dim>]
[B<--paperwidth> I<dim>]
[B<--photoheight> I<dim>]
[B<--photowidth> I<dim>]
[B<--resolution> I<dots/unit>]
[B<--verbose>]
[B<--version>]
 I<image_file>

=head1 DESCRIPTION

B<passport_photo> takes an image file containing a single picture
with suitable aspect (height to width) ratio and creates an image
for a specified printer's paper size and resolution containing as many copies of
the original image as will fit, with each scaled to a
precise size on the paper.  This permits economically creating
passport and other identification photos which conform to the size
requirements of those requesting them.  The program can also be
used to print multiple copies of other kinds of pictures on a page,
for example, snaps of the family for all the in-laws and outlaws.

B<passport_photo> is a Perl program which assembles images using
components of the Netpbm image processing toolkit; you must have
these utilities installed on your computer in order to use it.  While
these programs are usually associated with Unix (GNU/Linux, etc.)
systems, B<passport_photo> works fine on Windows systems with Cygwin
(B<http://www.cygwin.com/>) and the requisite programs and libraries
installed.

To obtain the highest quality results, start with a digital camera
image or scan of a photographic print made with the highest
reasonable resolution--when scanning a photo there's no point setting
the resolution so high you can see film grain and paper fibres.
The picture should be taken or cropped to "portrait mode" (greater
height than width) with the same aspect ratio as that required for
the individual photos.  For example, if photos of the European
standard of 45 millimetres high and 35 millimetres wide (aspect
ratio 1.286) are being made, and your 5 megapixel digital camera
produces images of 2592 by 1944 pixels (aspect ratio 1.333), you
should take the photo in portrait mode filling the full 1944 pixel
width, then crop the image with the face at the desired position
to a height of 2500 pixels to match the aspect ratio of the
printed photos.  Nothing terrible will happen if the aspect ratios
are slightly mismatched, but the image will be smaller due to the
need to scale it to fit within the required size.

=head1 OPTIONS

All options may be abbreviated to their shortest
unambiguous prefix.

=over 5

=item B<--black>

Use black borders between and around images instead of the
default white.  This is handy when you're using a printer
which doesn't fill the entire page ("bleed"), and you wish
to measure the actual print area so the individual photos will
be scaled to the desired size.  It's a poor idea to use
this option in other circumstances, as it wastes ink.

=item B<--copyright>

Display copyright information.

=item B<--grey> or B<--gray>

Convert a colour I<image_file> to grey scale.  Use this option when grey scale
(a.k.a. "black and white") photos are required or preferable, for example
when applying for a passport from a third world country most of whose
bearers have such photos (one place you I<never> want to stand
out is at the immigration desk), or a visa for Zeta Reticuli, where
such a complexion is I<de rigueur>.

=item B<--help>

Display how to call information.

=item B<--ifilter> I<program>

Specify the Netpbm input filter (for example B<pngtopnm>) used to
convert the input file to PPM format.  If the input file has an
extension of B<.jpg>, B<.jpeg>, B<.png>, B<.gif>, B<.tif>, B<.tiff>,
or B<.p>I<x>B<m>, the input filter will be automatically "guessed"
based on the file type if no B<--ifilter> is specified.  Note that
you can pass command line options to the input filter by quoting the
I<program> argument and specifying them after the filter
program name.

=item B<--inputfile> I<fname>

Specifies the name of the input file.  This is entirely equivalent
to specifying an I<image_file> argument.  In either case, a
specification of "B<->" causes input to be read from
standard input, in which case B<--ifilter> must be
specified.

=item B<--ofilter> I<program>

Specify the Netpbm output filter used to write the output file. The
default is B<pnmtojpeg>, and unless you I<really> know what you're
doing, you shouldn't change this as most services which print digital
camera files don't understand file formats other than JPEG.  Note
that you can pass command line options to the output filter by
quoting the I<program> argument and specifying them after the filter
program name.

=item B<--outputfile> I<fname>

Specify the name of the output file written by B<--ofilter>.
By default, output is written to standard output and may be
redirected to a file or piped to another program.

=item B<--paperheight> I<dim>

The height of the image produced by the printer is set to the
dimension I<dim>, where "height" is taken to mean the longer of the
dimensions of the image (hence "portrait mode").  Note that this
specifies the size of the image produced by the printer, I<not> the
paper size.  If the printer leaves a border around the image, you
should print an image with the B<--black> option and measure it to
determine the actual printed image size. The I<dim> argument is by
default in millimetres.  You may specify the dimension in other units
by appending a suffix of "B<cm>" for centimetres, "B<in>" for inches,
"B<pc>" for picas, or "B<pt>" for points.  Dimension values may be
decimal fractions, but are rounded to integral millimetres. These
considerations apply to I<dim> arguments of subsequent options.  The
default paper height is 150 millimetres (15 cm).

=item B<--paperwidth> I<dim>

The width of the image produced by the printer is set to I<dim>,
where "width" is the smaller of the dimensions of the image.  The
default paper width is 100 millimetres (10 cm).

=item B<--photoheight> I<dim>

Each individual photo will have height I<dim>, the larger of the two
image dimensions. The default photo height is 45 millimetres (4.5
cm).

=item B<--photowidth> I<dim>

Each individual photo will have width I<dim>, the smaller of the two
image dimensions.  The default photo width is 35 millimetres (3.5
cm).

=item B<--resolution> I<dots/unit>

The image will be prepared with a resolution of I<dots/unit>, where
the dimension is interpreted as a number of dots per unit of length
(millimetres if no unit is given). For example, "B<48/mm>" (the
default value) results in an image with a resolution of 48 dots per
millimetre, which is approximately the same as "B<1200/in>", 1200
dots per inch.

=item B<--verbose>

Generate verbose output to indicate what's going on, including a transcript
of all the Netpbm commands used to prepare the image.

=item B<--version>

Display version number.

=back

=head1 EXAMPLES

Produce a 10 by 15 centimetre sheet containing six 35 by 45 millimetre
passport photos from the image B<myshot.jpg>, writing the ready-to-print
48 dot per millimetre resolution image to JPEG file B<myprint.jpg>:

    perl passport_photo.pl myshot.jpg >myprint.jpg

Produce an 8 by 10 inch sheet at 1200 dots per inch resolution
containing twenty 2 by 2 inch photos as required for United
States visa applications.

    perl passport_photo.pl --paperheight 10in --paperwidth 8in
                           --photoheight 2in --photowidth 2in
                           --resolution 1200/in myshot.jpg >myvisa.jpg

=head1 FILES

Several intermediate temporary files are created in the current
directory, which are assembled into the final image.  These files
are rather large (about 48 megabytes with the default settings),
so make sure there's adequate space on the file system
containing the current directory.  If the program crashes or is
killed before completing, the intermediate files won't be
deleted.  They all have names which can be deleted by
a command such as "S<B<rm *_temp?.ppm>> as long as you don't
have any other files in the directory such a command would
inadvertently delete.

=head1 BUGS

Creating large sheets at high resolution takes I<lots> of memory and
substantial compute time.  If you run B<passport_photo> on a system
with less RAM than the uncompressed image size, forcing paging to
disc, it may take almost forever to generate an image.

One could add any number of fanciful length units to dimension
and resolution arguments: parsecs, angstroms, chains, nautical
miles, etc.  On the other hand, one should have better things
to do.

Yes, I've thought about calling this program B<ppmpassport> or
somesuch and submitting it as a Netpbm component.  But somebody
has to draw the line somewhere against accretion of special-purpose
kludges into Netpbm, and here's where I draw it.  If somebody
I<does> add this, they should remove the input file filtering
mechanism and require the input be piped to standard input
in PPM format.

Please report bugs to B<bugs@fourmilab.ch>, indicating the version
numbers of B<passport_photo>, Perl, and Netpbm installed on your
system.

=head1 AUTHOR

John Walker
(B<http://www.fourmilab.ch/>)

=head1 SEE ALSO

    http://netpbm.sourceforge.net/
    http://www.perl.org/

=head1 VERSION

This is B<passport_photo> version 1.1, released on January 29th, 2020.
The current version of this program is always posted at:

http://www.fourmilab.ch/netpbm/passport_photo/

=head1 COPYRIGHT

This program is in the public domain.

=cut

    use strict;
    use warnings;

    use Getopt::Long;

    my $version = '1.1';

    #   Size of print paper in millimetres
    my @paperSize = (100, 150);

    #   Resolution in dots per millimetre
    my $dotsPerMm = 48;

    #   Name of source photo
    my $photoName = '';

    #   Output file name
    my $ofileName = '';

    #   Filter to convert source photo image to PPM
    my $photoFilter = '';

    #   Convert image to grey scale ?
    my $grey = 0;

    #   Size of passport photos in millimetres
    my @photoSize = (35, 45);           # This is the current European default

    #   Output filter to prepare image to print
    my $ofilter = "pnmtojpeg";

    #   Add black border to image for measurement
    my $border = 0;

    #   Show verbose processing information ?
    my $verbose = 0;

    #   Units we recognise.  Each is given in terms of millimetres per unit
    my %units = (
                  'mm', 1,              # millimetres
                  'cm', 10,             # centimetres
                  'in', 25.4,           # inches
                  'pc', 4.2175176,      # picas
                  'pt', 0.3514598       # points
                );

    #   Process command line options

    my $ook = GetOptions(
               "black" => \$border,
               "copyright" => \&infoOption,
               "gray" => \$grey,
               "grey" => \$grey,
               "help" => \&infoOption,
               "ifilter=s" => \$photoFilter,
               "inputfile=s" => \$photoName,
               "ofilter=s" => \$ofilter,
               "outputfile=s" => \$ofileName,
               "paperheight=s" => \$paperSize[1],
               "paperwidth=s" => \$paperSize[0],
               "photoheight=s" => \$photoSize[1],
               "photowidth=s" => \$photoSize[0],
               "resolution=s" => \$dotsPerMm,
               "verbose" => \$verbose,
               "version" => \&infoOption
               );

    if (!$ook) {
        exit(1);            # Command line syntax error
    }

    #   Parse dimension specifications

    @paperSize = (dimension($paperSize[0]), dimension($paperSize[1]));
    @photoSize = (dimension($photoSize[0]), dimension($photoSize[1]));

    #   Parse resolution specification

    $dotsPerMm = resolution($dotsPerMm);

    #   Set up grey scale filter, if requested

    if ($grey) {
        $grey = "| ppmtopgm";
    } else {
        $grey = '';
    }

    #   Verify that input file name was specified and that it exists

    if ($photoName eq '') {
        if ($#ARGV >= 0) {
            $photoName = shift();
        } else {
            print(STDERR "No input file specified.\n");
            exit(1);
        }
    }
    if (($photoName ne '-') && !(-f $photoName)) {
        print(STDERR "Specified input file $photoName does not exist.\n");
        exit(1);
    }

    #   Check for extraneous non-option command line arguments

    if ($#ARGV > -1) {
        print(STDERR "Extraneous command line arguments ignored.\n");
    }

    #   If an output file name was specified, create a redirect command
    #   to send output there.  Otherwise, output will be sent to standard
    #   output.

    if ($ofileName ne '') {
        $ofileName = ">$ofileName";
    }

    #   Prepare border colour specifications

    my ($borderspec, $borderrgb) = $border ? ('-black', 'rgbi:0/0/0') : ('-white', 'rgbi:1/1/1');

    #   Compute how many photos will fit on the paper

    my @gridSize = (int($paperSize[0] / $photoSize[0]),
                    int($paperSize[1] / $photoSize[1]));

    print(STDERR "Paper size $paperSize[0]mm by $paperSize[1]mm.\n") if $verbose;
    print(STDERR "Photo size $photoSize[0]mm by $photoSize[1]mm.\n") if $verbose;
    print(STDERR "Laying out $gridSize[0] across, $gridSize[1] down.\n") if $verbose;

    #   Choose input filter based on input file extension

    if ($photoFilter eq '') {
        chooseFilter($photoName);
    }
    print(STDERR "Input filter command: \"$photoFilter\".\n") if $verbose;

    #   Create the prototype photo, scaled to the specified
    #   resolution and fit within the desired extent box.

    my $c;

    my ($px, $py) = ($photoSize[0] * $dotsPerMm, $photoSize[1] * $dotsPerMm);
    my $pid = $$;
    my $tfn1 = $$ . "_temp1.ppm";
    my $pn = ($photoName eq '-') ? '' : $photoName;
    $c = "$photoFilter $pn $grey | pnmscale -xysize $px $py >$tfn1";
    run($c);

    #   Determine size of scaled photo

    $c = `pnmfile $tfn1`;
    $c =~ m/,\s+(\d+)\s+by\s+(\d+)\s/;
    my ($imx, $imy) = ($1, $2);

    print(STDERR "Scaled image is $imx by $imy pixels.\n") if $verbose;

    #   Calculate size of strut between images

    my @margin = (int((($dotsPerMm * $paperSize[0]) - ($gridSize[0] * $imx)) / ($gridSize[0] + 1)),
                  int((($dotsPerMm * $paperSize[1]) - ($gridSize[1] * $imy)) / ($gridSize[1] + 1)));

    print(STDERR "Margin between images is $margin[0] by $margin[1].\n") if $verbose;

    #   Create vertical strut between images.  If images exactly
    #   fill the paper width, no strut is required.

    my $vstrut = "";
    if ($margin[0] > 0) {
        $vstrut = $$ . "_temp2.ppm";
        run("ppmmake $borderrgb $margin[0] $imy >$vstrut");
    }

    #   Assemble a horizontal row of images

    my $rowimg = $$ . "_temp3.ppm";
    my $j = 0;
    $c = "pnmcat $borderspec -leftright";
    for (my $i = 0; $i < $gridSize[0]; $i++) {
        $c .= " $vstrut $tfn1";
        $j += $margin[0] + $imx;
    }
    $j += $margin[0];
    print(STDERR "Concatenated width is $j pixels.\n") if $verbose;

    #   Test for round-off in width, create and concatenate strut if necessary

    if ($j < ($dotsPerMm * $paperSize[0])) {
        my $round_x_strut = $$ . "_temp6.ppm";
        my $rw = ($dotsPerMm * $paperSize[0]) - $j;
        run("ppmmake $borderrgb $rw $imy >$round_x_strut");
        $c .= " $round_x_strut";
    }

    $c .= " $vstrut >$rowimg";
    run($c);

    #   Create horizontal strut between rows of images.  If images
    #   exactly fill the paper width, no strut is required.

    my $hstrut = "";
    my $pixx = $paperSize[0] * $dotsPerMm;
    if ($margin[1] > 0) {
        $hstrut = $$ . "_temp4.ppm";
        run("ppmmake $borderrgb $pixx $margin[1] >$hstrut");
    }

    #   Assemble rows and struts into final image

    $c = "pnmcat $borderspec -topbottom";
    $j = 0;
    for (my $i = 0; $i < $gridSize[1]; $i++) {
        $c .= " $hstrut $rowimg";
        $j += $margin[1] + $imy;
    }
    $j += $margin[1];
    print(STDERR "Concatenated height is $j pixels.\n") if $verbose;

    #   Test for round-off in height, create and concatenate strut if necessary

    if ($j < ($dotsPerMm * $paperSize[1])) {
        my $round_y_strut = $$ . "_temp7.ppm";
        my $rw = ($dotsPerMm * $paperSize[1]) - $j;
        run("ppmmake $borderrgb $pixx $rw >$round_y_strut");
        $c .= " $round_y_strut";
    }

    $c .= " $hstrut | pnmflip -rotate90 | $ofilter $ofileName";
    run($c);

    #   Clean up intermediate files

    run("rm -f " . $$ . "_temp?.ppm");

    #   Echo a command to standard output and run it with system()

    sub run {
        my $cmd = shift();

        print(STDERR "$cmd\n") if $verbose;
        system($cmd);
    }

    #   Set input filter command based on input file's extension

    sub chooseFilter {
        my $fname = shift();

        if ($fname =~ m/\.jpe?g/i) {
            $photoFilter = 'jpegtopnm';
        } elsif ($fname =~ m/\.png/i) {
            $photoFilter = 'pngtopnm';
        } elsif ($fname =~ m/\.gif/i) {
            $photoFilter = 'giftopnm';
        } elsif ($fname =~ m/\.tiff?/i) {
            $photoFilter = 'tifftopnm';
        } elsif ($fname =~ m/\.p[abgp]m/i) {
            $photoFilter = 'cat';
        } else {
            die("No input filter defined for file: $fname");
        }
    }

    #   Parse dimension specification and return size in millimetres

    sub dimension {
        my $d = shift;
        my $n;

        if ($d =~ m/([\d\.]+)(\w+)?$/) {
            my $unit;

            ($n, $unit) = ($1, $2);
            if (!(defined $unit)) {
                $unit = 'mm';
            }
            if (!(defined $units{$unit})) {
                die("Invalid unit \'$unit\' in dimension specification: \"$d\".\n");
            }
            $n *= $units{$unit};
        } else {
            die("Invalid dimension specification: \"$d\".\n");
        }
#print(STDERR "Dimension $d = $n mm\n");
        return int($n);
    }

    #   Parse resolution specification and return dots per millimetre

    sub resolution {
        my $d = shift;
        my $n;

        if ($d =~ m/([\d\.]+)\/?(\w+)?$/) {
            my $unit;

            ($n, $unit) = ($1, $2);
            if (!(defined $unit)) {
                $unit = 'mm';
            }
            if (!(defined $units{$unit})) {
                die("Invalid unit \'$unit\' in resolution specification: \"$d\".\n");
            }
            $n /= $units{$unit};
        } else {
            die("Invalid resolution specification: \"$d\".\n");
        }

        return int($n + 0.5);
    }

    #   Process information request options

    sub infoOption {
        my ($opt) = @_;

#print("=$opt=\n");
        if ($opt eq'copyright') {
            print("This program is in the public domain.\n");
            exit(0);
        }

        if ($opt eq 'help') {
            print << "EOD";
usage: passport_photo [ options ] image_file

Options:
  --black                  Use black border around pictures
  --copyright              Display copyright information
  --help                   Display this message
  --ifilter prog           Use graphics file to PPM filter for input file
  --inputfile              Input file name (alternative to image_file argument)
  --ofilter prog           Output file filter (default pnmtojpeg)
  --outputfile fname       Output file name (default standard output)
  --paperheight dim        Print paper height, portrait mode (default 150mm)
  --paperwidth dim         Print paper width, portrait mode (default 100mm)
  --photoheight dim        Passport photo height (default 45mm)
  --photowidth dim         Passport photo width (default 35mm)
  --resolution dpmm        Resolution in dots per millimetre (default 48)
  --verbose                Display verbose status information
  --version                Display version number

Please report bugs to bugs\@fourmilab.ch; include "passport_photo"
in the Subject line.
EOD

            exit(0);
        }

        if ($opt eq 'version') {
            print("$version\n");
            exit(0);
        }

    }
